# coding=utf-8
from __future__ import print_function

import pandas as pd

from suanpan import path

try:
    from io import StringIO
except:
    from StringIO import StringIO


def load(file, *args, **kwargs):
    kwargs.setdefault("index_col", 0)
    return pd.read_csv(file, *args, **kwargs)


def loads(string, *args, **kwargs):
    return pd.read_csv(StringIO(string), *args, **kwargs)


def dump(dataframe, file, *args, **kwargs):
    path.safeMkdirsForFile(file)
    dataframe.to_csv(file, *args, **kwargs)
    return file


def dumps(dataframe, *args, **kwargs):
    output = StringIO()
    dataframe.to_csv(output, *args, **kwargs)
    return output.getvalue()



# coding=utf-8
from __future__ import print_function

from pyspark2pmml import PMMLBuilder
from pyspark.ml import PipelineModel

from suanpan import objects
from suanpan.arguments import *  # pylint: disable-msg=w0614
from suanpan.spark import db, io
from suanpan.utils import json


class Table(Arg):
    def __init__(self, key, table, partition, sortColumns=None):
        super(Table, self).__init__(key)
        sortColumns = sortColumns or "{}SortColumns".format(table)
        self.table = String(key=table, required=True)
        self.partition = String(key=partition)
        self.sortColumns = ListOfString(key=sortColumns, default=[])

    def addParserArguments(self, parser):
        self.table.addParserArguments(parser)
        self.partition.addParserArguments(parser)
        self.sortColumns.addParserArguments(parser)

    def load(self, args):
        self.table.load(args)
        self.partition.load(args)
        self.sortColumns.load(args)
        self.value = dict(
            table=self.table.value,
            partition=self.partition.value,
            sortColumns=self.sortColumns.value,
        )

    def format(self, context):
        self.value = db.readTable(context.spark, self.table.value, self.partition.value)
        self.value = self.value.sort(self.sortColumns.value)
        return self.value

    def save(self, context, result):
        data = result.value
        db.writeTable(context.spark, self.table.value, data)
        self.logSaved(self.table.value)


Table = Table


class SparkModel(Arg):
    def __init__(self, key, **kwargs):
        kwargs.update(required=True)
        super(SparkModel, self).__init__(key, **kwargs)

    def format(self, context):
        self.value = PipelineModel.load(io.getStoragePath(context.spark, self.value))
        return self.value

    def save(self, context, result):
        spark = context.spark

        modelPath = io.getStoragePath(spark, self.value)
        pmmlPath = modelPath + "/pmml"

        result.value.model.write().overwrite().save(modelPath)
        with io.open(spark, pmmlPath, mode="w") as file:
            pmmlBuilder = PMMLBuilder(
                spark.sparkContext, result.value.data, result.value.model
            ).putOption(result.value.classifier, "compact", True)
            pmml = pmmlBuilder.buildByteArray()
            file.write(pmml)


class PmmlModel(Arg):
    def __init__(self, key, **kwargs):
        kwargs.update(required=True)
        super(PmmlModel, self).__init__(key, **kwargs)

    def format(self, context):
        spark = context.spark

        pmmlPath = self.pmml_path(spark)
        with io.open(spark, pmmlPath, mode="r") as file:
            self.value = file.read()

        return self.value

    def pmml_path(self, spark):
        modelPath = io.getStoragePath(spark, self.value)
        pmmlPath = modelPath + "/pmml"
        return pmmlPath


class OdpsTable(Arg):
    def __init__(
        self,
        key,
        accessId,
        accessKey,
        odpsUrl,
        tunnelUrl,
        project,
        table,
        partition,
        overwrite,
        numPartitions,
    ):
        super(OdpsTable, self).__init__(key)
        self.accessId = String(key=accessId, required=True)
        self.accessKey = String(key=accessKey, required=True)
        self.odpsUrl = String(key=odpsUrl, required=True)
        self.tunnelUrl = String(key=tunnelUrl, required=True)
        self.project = String(key=project, required=True)
        self.table = String(key=table, required=True)
        self.partition = String(key=partition)
        self.overwrite = Bool(key=overwrite, default=False)
        self.numPartitions = Int(key=numPartitions, default=2)

    def addParserArguments(self, parser):
        self.accessId.addParserArguments(parser)
        self.accessKey.addParserArguments(parser)
        self.odpsUrl.addParserArguments(parser)
        self.tunnelUrl.addParserArguments(parser)
        self.project.addParserArguments(parser)
        self.table.addParserArguments(parser)
        self.partition.addParserArguments(parser)
        self.overwrite.addParserArguments(parser)
        self.numPartitions.addParserArguments(parser)

    def load(self, args):
        self.accessId.load(args)
        self.accessKey.load(args)
        self.odpsUrl.load(args)
        self.tunnelUrl.load(args)
        self.project.load(args)
        self.table.load(args)
        self.partition.load(args)
        self.overwrite.load(args)
        self.numPartitions.load(args)
        self.value = dict(
            accessId=self.accessId.value,
            accessKey=self.accessKey.value,
            odpsUrl=self.odpsUrl.value,
            tunnelUrl=self.tunnelUrl.value,
            table=self.table.value,
            partition=self.partition.value,
            overwrite=self.overwrite.value,
            numPartitions=self.numPartitions.value,
        )

    def format(self, context):
        self.value = db.readOdpsTable(
            context.spark,
            accessId=self.accessId.value,
            accessKey=self.accessKey.value,
            odpsUrl=self.odpsUrl.value,
            tunnelUrl=self.tunnelUrl.value,
            project=self.project.value,
            table=self.table.value,
            partition=self.partition.value,
            numPartitions=self.numPartitions.value,
        )
        return self.value

    def save(self, context, result):
        db.writeOdpsTable(
            context.spark,
            accessId=self.accessId.value,
            accessKey=self.accessKey.value,
            odpsUrl=self.odpsUrl.value,
            tunnelUrl=self.tunnelUrl.value,
            project=self.project.value,
            table=self.table.value,
            data=result.value,
            partition=self.partition.value,
            overwrite=self.overwrite.value,
        )


class Visual(Arg):
    def __init__(self, key, **kwargs):
        kwargs.update(required=True)
        super(Visual, self).__init__(key, **kwargs)

    def save(self, context, result):
        spark = context.spark

        visualPath = io.getStoragePath(spark, self.value) + "/part-00000"
        with io.open(spark, visualPath, mode="w") as file:
            file.write(result.value)


class JsonFile(Arg):
    def format(self, context):
        spark = context.spark

        jsonPath = self.filePath(spark)
        with io.open(spark, jsonPath, mode="r") as file:
            self.value = json.load(file)
            return self.value

    def save(self, context, result):
        spark = context.spark

        data = (
            result.value.to_dict()
            if isinstance(result.value, objects.Context)
            else result.value
        )
        jsonPath = self.filePath(spark)
        with io.open(spark, jsonPath, mode="w") as file:
            json.dump(data, file)

    def filePath(self, spark):
        dataPath = io.getStoragePath(spark, self.value)
        jsonPath = "{}/data.json".format(dataPath)
        return jsonPath



# coding=utf-8
from __future__ import print_function

from suanpan import arguments as base
from suanpan.dw import arguments as dw
from suanpan.storage import arguments as storage

Int = base.Int
String = base.String
Float = base.Float
Bool = base.Bool
List = base.List
ListOfString = base.ListOfString
ListOfInt = base.ListOfInt
ListOfFloat = base.ListOfFloat
ListOfBool = base.ListOfBool
File = storage.File
Folder = storage.Folder
Data = storage.Data
Json = storage.Json
Csv = storage.Csv
Npy = storage.Npy
Model = storage.Model
H5Model = storage.H5Model
Checkpoint = storage.Checkpoint
JsonModel = storage.JsonModel
Table = dw.Table



# coding=utf-8
from __future__ import division, print_function

import os
import tempfile
import zipfile

from suanpan import path as spath
from suanpan.log import logger
from suanpan.storage.objects import Storage


class LocalStorage(Storage):
    def __init__(self, localTempStore=tempfile.gettempdir(), **kwargs):
        super(LocalStorage, self).__init__(
            delimiter=os.sep, tempStore=localTempStore, **kwargs
        )

    def download(self, name, path):
        logger.info(
            "Downloading: {} -> {} - ({})Nothing to do!".format(
                self.storageUrl(name), path, self.name
            )
        )
        return self.storagePathJoin(self.tempStore, path)

    def upload(self, name, path):
        logger.info(
            "Uploading: {} -> {} - ({})Nothing to do!".format(
                self.storageUrl(name), path, self.name
            )
        )
        return path

    def copy(self, path, dist):
        pathUrl = self.storageUrl(path)
        distUrl = self.storageUrl(dist)
        logger.info("Copying: {} -> {}".format(pathUrl, distUrl))
        _path = self.storagePathJoin(self.tempStore, path)
        _dist = self.storagePathJoin(self.tempStore, dist)
        spath.copy(_path, _dist)
        logger.info("Copied: {} -> {}".format(pathUrl, distUrl))
        return dist

    def remove(self, path):
        spath.remove(self.storagePathJoin(self.tempStore, path))
        logger.info("Removed: {}".format(self.storageUrl(path)))
        return path

    def walk(self, folderName):
        root = self.storagePathJoin(self.tempStore, folderName)
        return os.walk(root)

    def listAll(self, folderName):
        root = self.storagePathJoin(self.tempStore, folderName)
        return (self.storagePathJoin(root, p) for p in os.listdir(root))

    def listFolders(self, folderName):
        return (p for p in self.listAll(folderName) if os.path.isdir(p))

    def listFiles(self, folderName):
        return (p for p in self.listAll(folderName) if os.path.isfile(p))

    def isFolder(self, folderName):
        folder = self.storagePathJoin(self.tempStore, folderName)
        return os.path.isdir(folder)

    def isFile(self, objectName):
        file = self.storagePathJoin(self.tempStore, objectName)
        return os.path.isfile(file)

    def storagePathJoin(self, *paths):
        return self.localPathJoin(*paths)

    def storageRelativePath(self, path, base):
        return self.localRelativePath(path, base)

    def storageUrl(self, path):
        return "file://" + self.storagePathJoin(self.tempStore, path)



# coding=utf-8
from __future__ import print_function

import argparse
import contextlib
import itertools
import traceback
from collections import defaultdict

from suanpan import interfaces, objects
from suanpan.log import logger


class Arguments(objects.Context):
    pass


class Result(objects.Context):
    def __init__(self, value):
        super(Result, self).__init__(value=value)


class Component(interfaces.HasArguments):
    def __init__(self, funcOrComponent):
        if isinstance(funcOrComponent, Component):
            self.runFunc = funcOrComponent.runFunc
            self.arguments = funcOrComponent.arguments
        else:
            self.runFunc = funcOrComponent
            self.arguments = defaultdict(list)

    def __call__(self, *arg, **kwargs):
        try:
            self.run(*arg, **kwargs)
        except Exception as e:
            logger.error(traceback.format_exc())
            raise e

    @property
    def name(self):
        return self.runFunc.__name__

    def run(self, *arg, **kwargs):
        logger.info("Starting...")
        context = self.init()
        results = self.runFunc(context)
        self.save(context, results)
        logger.info("Done.")

    def init(self):
        restArgs = self.getArgList()
        globalArgs, restArgs = self.loadGlobalArguments(restArgs=restArgs)
        self.initBase(globalArgs)
        context = self._getContext(globalArgs)
        args, restArgs = self.loadComponentArguments(context, restArgs=restArgs)
        setattr(context, "args", args)
        return context

    def initBase(self, args):
        pass

    def save(self, context, results):
        outputs = self.saveOutputs(context, results)
        self._closeContext()
        return outputs

    @contextlib.contextmanager
    def context(self, args=None):
        yield objects.Context()

    def _getContext(self, *args, **kwargs):
        self.contextManager = self.context(*args, **kwargs)
        return next(self.contextManager.gen)  # pylint: disable-msg=e1101

    def _closeContext(self):
        try:
            next(self.contextManager.gen)  # pylint: disable-msg=e1101
        except StopIteration:
            pass

    def loadGlobalArguments(self, restArgs=None, **kwargs):
        logger.info("Loading Global Arguments:")
        arguments, restArgs = super(Component, self).loadGlobalArguments(
            restArgs=restArgs, **kwargs
        )
        args = {arg.key: arg.value for arg in arguments}
        return Arguments(**args), restArgs

    def loadComponentArguments(self, context, restArgs=None):
        logger.info("Loading Component Arguments:")
        _, restArgs = self.loadFormatArguments(
            context, restArgs=restArgs, exclude="outputs"
        )
        _, restArgs = self.loadCleanArguments(
            context, restArgs=restArgs, include="outputs"
        )
        arguments = {
            k: {arg.key: arg.value for arg in v if arg.isSet}
            for k, v in self.arguments.items()
        }
        arguments.update(
            {arg.key: arg.value for arg in itertools.chain(*self.arguments.values())}
        )
        return Arguments(**arguments), restArgs

    def getArguments(self, include=None, exclude=None):
        includes = set(self.arguments.keys() if not include else self._list(include))
        excludes = set([] if not exclude else self._list(exclude))
        includes = includes - excludes
        argumentsLists = [self.arguments[c] for c in includes]
        return list(itertools.chain(*argumentsLists))

    def saveMutipleOutputs(self, context, outputs, results):
        if isinstance(results, tuple) or isinstance(results, list):
            results = (Result(result) for result in results)
            return self.saveArguments(outputs, results)
        elif isinstance(results, dict):
            outputs, results = zip(
                *[
                    (argument, Result(results[argument.key]))
                    for argument in outputs
                    if results.get(argument.key) is not None
                ]
            )
            return self.saveArguments(outputs, results)
        else:
            raise Exception("Incorrect results: {}".format(results))

    def saveOneOutput(self, context, output, results):
        result = (
            Result(results[output.key])
            if isinstance(results, dict) and output.key in results
            else Result(results)
        )
        return {output.key: output.save(context, result)}

    def saveOutputs(self, context, results):
        logger.info("Saving...")
        outputs = self.getArguments(include="outputs")
        if len(outputs) > 1:
            return self.saveMutipleOutputs(context, outputs, results)
        elif len(outputs) == 1:
            return self.saveOneOutput(context, outputs[0], results)

    def addArgument(self, arg, argtype="args", reverse=True):
        if reverse:
            self.arguments[argtype].insert(0, arg)
        else:
            self.arguments[argtype].append(arg)

    @classmethod
    def arg(cls, argument, argtype="args", reverse=True):
        def _dec(funcOrComponent):
            funcOrComponent = (
                funcOrComponent
                if isinstance(funcOrComponent, cls)
                else cls(funcOrComponent)
            )
            funcOrComponent.addArgument(argument, argtype=argtype, reverse=reverse)
            return funcOrComponent

        return _dec

    @classmethod
    def input(cls, *args, **kwargs):
        kwargs.update(argtype="inputs")
        return cls.arg(*args, **kwargs)

    @classmethod
    def output(cls, *args, **kwargs):
        kwargs.update(argtype="outputs")
        return cls.arg(*args, **kwargs)

    @classmethod
    def param(cls, *args, **kwargs):
        kwargs.update(argtype="params")
        return cls.arg(*args, **kwargs)

    @classmethod
    def column(cls, *args, **kwargs):
        kwargs.update(argtype="columns")
        return cls.arg(*args, **kwargs)

    def _list(self, params=None):
        return [params] if isinstance(params, str) else list(params)



# coding=utf-8
from __future__ import print_function

import argparse
import contextlib
import itertools
import time
import traceback
import uuid

from suanpan import arguments as baseargs
from suanpan import g, interfaces, objects
from suanpan.arguments import Bool, Int, String
from suanpan.components import Component, Arguments
from suanpan.dw import dw
from suanpan.log import logger
from suanpan.mq import mq
from suanpan.mstorage import mstorage
from suanpan.storage import storage
from suanpan.utils import json


class Handler(Component):
    def __call__(self, steamObj, message, *arg, **kwargs):
        return self.run(steamObj, message, *arg, **kwargs)

    def run(self, steamObj, message, *arg, **kwargs):
        handlerContext = self.init(message)
        results = self.runFunc(steamObj, handlerContext)
        return self.save(handlerContext, results)

    def init(self, message):
        restArgs = self.getArgList(message)
        globalArgs, restArgs = self.loadGlobalArguments(restArgs=restArgs)
        self.initBase(globalArgs)
        handlerContext = self._getContext(message)
        args, restArgs = self.loadComponentArguments(handlerContext, restArgs=restArgs)
        setattr(handlerContext, "args", args)
        return handlerContext

    @contextlib.contextmanager
    def context(self, message):
        yield objects.Context(message=message)

    def getArgList(self, message):
        inputArguments = itertools.chain(
            *[
                ["--{}".format(arg.key), message.get("in{}".format(i + 1))]
                for i, arg in enumerate(self.getArguments(include="inputs"))
                if message.get("in{}".format(i + 1)) is not None
            ]
        )
        outputArguments = itertools.chain(
            *[
                [
                    "--{}".format(arg.key),
                    self.getOutputTmpValue(message, "out{}".format(i + 1)),
                ]
                for i, arg in enumerate(self.getArguments(include="outputs"))
            ]
        )
        return list(itertools.chain(inputArguments, outputArguments))

    def saveOutputs(self, context, results):
        if results is not None:
            outputs = super(Handler, self).saveOutputs(context, results)
            outputs = self.formatAsOuts(outputs)
            outputs = self.stringifyOuts(outputs)
            return outputs

    def formatAsOuts(self, results):
        return {
            "out{}".format(i + 1): results[arg.key]
            for i, arg in enumerate(self.getArguments(include="outputs"))
            if results[arg.key] is not None
        }

    def stringifyOuts(self, outs):
        return {k: str(v) for k, v in outs.items()}

    def shortenRequestID(self, requestID):
        return requestID.replace("-", "")

    def getOutputTmpValue(self, message, output):
        shortRequestID = self.shortenRequestID(message["id"])
        return baseargs.DEFAULT_ARG_VALUE_DELIMITER.join(
            [
                g.USER_ID,  # pylint: disable=no-member
                g.APP_ID,  # pylint: disable=no-member
                g.NODE_ID,  # pylint: disable=no-member
                output,
                shortRequestID,
            ]
        )


class Stream(interfaces.HasBaseServices, interfaces.HasLogger, interfaces.HasDevMode):

    DEFAULT_STREAM_CALL = "call"
    STREAM_ARGUMENTS = [
        String("stream-user-id", required=True),
        String("stream-app-id", required=True),
        String("stream-node-id", required=True),
        String("stream-node-group", default="default"),
        String("stream-recv-queue", required=True),
        String("stream-send-queue", required=True),
    ]

    def __init__(self):
        super(Stream, self).__init__()
        restArgs = self.getArgList()
        arguments, restArgs = self.loadGlobalArguments(restArgs=restArgs)
        self.args = Arguments(
            **{argument.key: argument.value for argument in arguments}
        )
        self.setOptions(self.args)
        self.setGlobals(self.args)
        self.setBaseServices(self.args)
        self.afterInit()

    def getGlobalArguments(self, *args, **kwargs):
        arguments = super(Stream, self).getGlobalArguments(*args, **kwargs)
        return arguments + self.STREAM_ARGUMENTS

    def generateRequestId(self):
        return uuid.uuid4().hex

    def formatMessage(self, message, msg, costTime=None):
        msgs = [message["id"], message.get("type", self.DEFAULT_STREAM_CALL), msg]
        if costTime is not None:
            msgs.insert(-1, "{}s".format(costTime))
        return " - ".join(msgs)

    def streamCall(self, message):
        logger.info(self.formatMessage(message, msg="Start"))
        startTime = time.time()
        try:
            callFunction = self.getCallFunction(message)
            outputs = callFunction(self, message) or {}
            endTime = time.time()
            costTime = round(endTime - startTime, 3)
            logger.info(self.formatMessage(message, msg="Done", costTime=costTime))
            if outputs:
                self.sendSuccessMessage(message, outputs)
        except:
            tracebackInfo = traceback.format_exc()
            endTime = time.time()
            costTime = round(endTime - startTime, 3)
            logger.error(
                self.formatMessage(message, msg=tracebackInfo, costTime=costTime)
            )
            self.sendFailureMessage(message, tracebackInfo)

    def getCallFunction(self, message):
        streamCall = message.get("type", self.DEFAULT_STREAM_CALL)
        callFunction = getattr(self, streamCall, None)
        if not callFunction:
            raise Exception("Unknown stream call: {}.{}".format(self.name, streamCall))
        return callFunction

    def call(self, message, *args):
        raise NotImplementedError("Method not implemented!")

    def start(self):
        for message in self.subscribe():
            self.streamCall(message["data"])

    def setDefaultMessageType(self, message):
        message["data"].setdefault("type", self.DEFAULT_STREAM_CALL)
        return message

    def getMessageExtraData(self, message):
        extra = message["data"].get("extra")
        extra = json.loads(extra) if extra else {}
        message["data"].update(extra=extra)
        return message

    def setOptions(self, args):
        self.options = self.defaultArgumentsFormat(args, self.STREAM_ARGUMENTS)
        return self.options

    def setGlobals(self, args):
        g.set("USER_ID", args.stream_user_id)
        g.set("APP_ID", args.stream_app_id)
        g.set("NODE_ID", args.stream_node_id)
        g.set("NODE_GROUP", args.stream_node_group)
        return g

    def afterInit(self):
        pass

    def createQueues(self, force=False):
        mq.createQueue(self.options["recvQueue"], force=force)
        mq.createQueue(self.options["sendQueue"], force=force)

    def subscribe(self, **kwargs):
        for message in mq.subscribeQueue(
            self.options["recvQueue"],
            group=self.options["nodeGroup"],
            consumer=self.options["nodeId"],
            **kwargs
        ):
            message = self.setDefaultMessageType(message)
            message = self.getMessageExtraData(message)
            yield message

    def recv(self, **kwargs):
        return mq.recvMessages(
            self.options["recvQueue"],
            group=self.options["nodeId"],
            consumer=self.name,
            **kwargs
        )

    def send(self, message, data, queue=None, extra=None):
        queue = queue or self.options["sendQueue"]
        message["extra"].update(extra or {})
        data = {
            "node_id": self.options["nodeId"],
            "request_id": message["id"],
            "type": message.get("type", self.DEFAULT_STREAM_CALL),
            "extra": json.dumps(message["extra"]),
            **data,
        }
        logger.debug("Send to `{}`: {}".format(queue, data))
        return mq.sendMessage(queue, data)

    def sendSuccessMessage(self, message, data, queue=None, extra=None):
        keys = ["out{}".format(i + 1) for i in range(5)]
        if not self.keysAllIn(data.keys(), keys):
            raise Exception("Success Message data only accept keys: {}".format(keys))
        data = {key: data.get(key) for key in keys if data.get(key) is not None}
        data.update(success="true")
        return self.send(message, data, queue=queue, extra=extra)

    def sendFailureMessage(self, message, msg, queue=None, extra=None):
        if not isinstance(msg, str):
            raise Exception("Failure Message msg only accept string")
        data = {"msg": msg, "success": "false"}
        return self.send(message, data, queue=queue, extra=extra)

    def sendMissionMessage(self, message, data, queue=None, extra=None):
        keys = ["in{}".format(i + 1) for i in range(5)]
        if not self.keysAllIn(data.keys(), keys):
            raise Exception("Mission Message data only accept keys: {}".format(keys))
        return self.send(message, data, queue=queue, extra=extra)

    def keysAllIn(self, keys, kset):
        return len(set(keys) - set(kset)) == 0



# coding=utf-8
from __future__ import print_function

import abc
import functools
import itertools
import os
import shutil
import tempfile
import zipfile

from suanpan import objects
from suanpan import path as spath
from suanpan.log import logger


class Storage(abc.ABC, objects.HasName):
    DEFAULT_IGNORE_KEYWORDS = {"__MACOSX", ".DS_Store"}
    CONTENT_MD5 = "Content-MD5"
    PBAR_FORMAT = "{l_bar}{bar}"

    def __init__(self, delimiter=os.sep, tempStore=tempfile.gettempdir(), **kwargs):
        self.delimiter = delimiter
        self.tempStore = tempStore

    @abc.abstractmethod
    def download(self, name, path, **kwargs):
        pass

    @abc.abstractmethod
    def upload(self, name, path, **kwargs):
        pass

    @abc.abstractmethod
    def copy(self, name, dist, **kwargs):
        pass

    @abc.abstractmethod
    def remove(self, name, **kwargs):
        pass

    @abc.abstractmethod
    def walk(self, folder, **kwargs):
        pass

    @abc.abstractmethod
    def listAll(self, folder, **kwargs):
        pass

    @abc.abstractmethod
    def listFolders(self, folder, **kwargs):
        pass

    @abc.abstractmethod
    def listFiles(self, folder, **kwargs):
        pass

    @abc.abstractmethod
    def isFolder(self, path, **kwargs):
        pass

    @abc.abstractmethod
    def isFile(self, path, **kwargs):
        pass

    def compress(self, zipFilePath, path, ignores=DEFAULT_IGNORE_KEYWORDS):
        compressFunc = self.compressFolder if os.path.isdir(path) else self.compressFile
        return compressFunc(zipFilePath, path)

    def compressFolder(self, zipFilePath, folderPath, ignores=DEFAULT_IGNORE_KEYWORDS):
        if folderPath in ignores:
            logger.info(
                "Ignore compressing folder: {} -> {}".format(folderPath, zipFilePath)
            )
            return zipFilePath

        logger.info("Compressing folder: {} -> {}".format(folderPath, zipFilePath))
        with zipfile.ZipFile(zipFilePath, "w") as zip:
            for root, _, files in os.walk(folderPath):
                for file in files:
                    filePath = os.path.join(root, file)
                    zip.write(
                        filePath, arcname=self.localRelativePath(filePath, folderPath)
                    )
        logger.info("Compressed folder: {} -> {}".format(folderPath, zipFilePath))
        return zipFilePath

    def compressFile(self, zipFilePath, filePath, ignores=DEFAULT_IGNORE_KEYWORDS):
        if filePath in ignores:
            logger.info(
                "Ignore compressing File: {} -> {}".format(filePath, zipFilePath)
            )
            return zipFilePath

        logger.info("Compressing File: {} -> {}".format(filePath, zipFilePath))
        with zipfile.ZipFile(zipFilePath, "w") as zip:
            _, filename = os.path.split(filePath)
            zip.write(filePath, arcname=filename)
        logger.info("Compressed File: {} -> {}".format(filePath, zipFilePath))
        return zipFilePath

    def extract(self, zipFilePath, distPath, ignores=DEFAULT_IGNORE_KEYWORDS):
        logger.info("Extracting zip: {} -> {}".format(zipFilePath, distPath))
        with zipfile.ZipFile(zipFilePath, "r") as zip:
            zip.extractall(distPath)
        self.removeIgnores(distPath, ignores=ignores)
        logger.info("Extracted zip: {} -> {}".format(zipFilePath, distPath))

    def getPathInTempStore(self, path, tempStore=None):
        tempStore = tempStore or self.tempStore
        return self.localPathJoin(tempStore, path)

    def completePath(self, path, delimiter=None):
        delimiter = delimiter or self.delimiter
        return path if path.endswith(delimiter) else path + delimiter

    def toLocalPath(self, path, delimiter=None):
        delimiter = delimiter or self.delimiter
        return path.replace(delimiter, os.sep)

    def toStoragePath(self, path, delimiter=None):
        delimiter = delimiter or self.delimiter
        return path.replace(os.sep, delimiter)

    def localPathJoin(self, *paths):
        path = os.path.join(*paths)
        return self.toLocalPath(path)

    def storagePathJoin(self, *paths):
        path = os.path.join(*[self.toLocalPath(path) for path in paths])
        return self.toStoragePath(path)

    def localRelativePath(self, path, base):
        return self._relativePath(path, base, delimiter=os.sep)

    def storageRelativePath(self, path, base, delimiter=None):
        delimiter = delimiter or self.delimiter
        return self._relativePath(path, base, delimiter=delimiter)

    def _relativePath(self, path, base, delimiter):
        base = base if base.endswith(delimiter) else base + delimiter
        return path[len(base) :] if path.startswith(base) else path

    @abc.abstractmethod
    def getStorageMd5(self, name, **kwargs):
        pass

    def getLocalMd5(self, path, **kwargs):
        return spath.md5(path) if os.path.isfile(path) else None

    def checkMd5(self, md5a, md5b, **kwargs):
        return md5a if md5a == md5b and md5a is not None else False

    @abc.abstractmethod
    def getStorageSize(self, name, **kwargs):
        pass

    def getLocalSize(self, path, **kwargs):
        return os.path.getsize(path)

    def storageUrl(self, path, **kwargs):
        return "storage://" + path

    def removeIgnores(self, path, ignores=None):
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS

        def _ignore(_path):
            spath.remove(_path)
            logger.info("Removed ignored: {}".format(_path))
            return _path

        def _getIgnores(path):
            for root, folders, files in os.walk(path):
                for folder in folders:
                    if folder in ignores:
                        yield os.path.join(root, folder)
                for file in files:
                    if file in ignores:
                        yield os.path.join(root, file)

        return [_ignore(_path) for _path in _getIgnores(path)]

    def pathSplit(self, path, delimiter=None):
        delimiter = delimiter or self.delimiter
        return path.rsplit(delimiter, 1)

    def pathSplitExt(self, path, extDelimiter="."):
        return path.rsplit(extDelimiter, 1)

    def pbarRunner(self, pbar, quantity=1):
        def _dec(runner):
            @functools.wraps(runner)
            def _runner(*args, **kwargs):
                result = runner(*args, **kwargs)
                pbar.update(quantity)
                return result

            return _runner

        return _dec



# coding=utf-8
from __future__ import print_function

import argparse
import os

from suanpan import objects
from suanpan.log import logger
from suanpan.utils import json

DEFAULT_ARG_VALUE_DELIMITER = "::"
DEFAULT_ACTUAL_ARG_VALUE_DELIMITER = "_"
DEFAULT_MAX_VALUE_LENGTH = 120


class Arg(objects.HasName):
    ARG_VALUE_DELIMITER = DEFAULT_ARG_VALUE_DELIMITER
    ACTUAL_ARG_VALUE_DELIMITER = DEFAULT_ACTUAL_ARG_VALUE_DELIMITER
    MAX_VALUE_LENGTH = DEFAULT_MAX_VALUE_LENGTH

    def __init__(self, key, **kwargs):
        if "default" in kwargs:
            kwargs.update(required=False)

        self.argkey = key
        self.key = self.fixGlobalKey(key)
        self.value = None
        self.type = kwargs.pop("type", str)
        self.required = kwargs.pop("required", False)
        self.default = kwargs.pop("default", None)

        self.kwargs = self.cleanParams(kwargs)
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def isSet(self):
        return self.required or self.default != self.value

    def addParserArguments(self, parser, required=None, default=None):
        required = self.required if required is None else required
        default = default or self.default
        return parser.add_argument(
            "--{}".format(self.argkey),
            type=self.typeDecorator(self.type),
            required=required,
            default=default,
            **self.kwargs
        )

    def addGlobalParserArguments(self, parser, required=None, default=None):
        default = os.environ.get(self.envKeyFormat(self.key), default or self.default)
        self.addParserArguments(parser, required=required, default=default)

    def typeDecorator(self, typeFunc):
        def _decorator(value):
            return getattr(self, "default", None) if value == "" else typeFunc(value)

        return _decorator

    def load(self, args):
        self.value = getattr(args, self.key)
        self.value = self.fixArgValue(self.value)
        self.logLoaded(self.value)
        return self.value

    def format(self, context):
        return self.value

    def clean(self, context):
        return self.value

    def save(self, context, result):
        self.logSaved(result.value)
        return result.value

    def cleanParams(self, params):
        return {k: v for k, v in params.items() if not k.startswith("_")}

    def logLoaded(self, value):
        logger.info(
            "({type}) {key} loaded: {value}".format(
                type=self.name,
                key=self.key,
                value=self.shortenFormat(value, maxlen=self.MAX_VALUE_LENGTH),
            )
        )

    def logSaved(self, value):
        logger.info(
            "({type}) {key} saved: {value}".format(
                type=self.name,
                key=self.key,
                value=self.shortenFormat(value, maxlen=self.MAX_VALUE_LENGTH),
            )
        )

    def fixArgValue(self, value):
        return value

    def fixGlobalKey(self, key):
        return key.replace("-", "_")

    def shortenFormat(self, value, maxlen=DEFAULT_MAX_VALUE_LENGTH):
        string = str(value)
        return string[: maxlen - 3] + "..." if len(string) > maxlen else string

    def envKeyFormat(self, key):
        return self.fixGlobalKey(key).upper()


class String(Arg):
    def __init__(self, key, **kwargs):
        super(String, self).__init__(key, type=str, **kwargs)


class Int(Arg):
    def __init__(self, key, **kwargs):
        super(Int, self).__init__(key, type=int, **kwargs)


class Float(Arg):
    def __init__(self, key, **kwargs):
        super(Float, self).__init__(key, type=float, **kwargs)


class Bool(Arg):
    def __init__(self, key, **kwargs):
        kwargs.update(default=False)
        super(Bool, self).__init__(key, type=type(self).str2bool, **kwargs)

    @classmethod
    def str2bool(cls, string):
        if string.lower() in ("yes", "true", "t", "y", "1"):
            return True
        elif string.lower() in ("no", "false", "f", "n", "0"):
            return False
        else:
            raise argparse.ArgumentTypeError("Boolean value expected.")


class List(Arg):
    def __init__(self, key, **kwargs):
        super(List, self).__init__(key, type=type(self).str2list, **kwargs)

    @classmethod
    def str2list(cls, string):
        try:
            return [cls.transform(i.strip()) for i in string.split(",") if i.strip()]
        except:
            raise argparse.ArgumentTypeError("{} value expected.".format(cls.__name__))

    @classmethod
    def transform(cls, item):
        return item


class ListOfString(List):
    pass


class ListOfInt(List):
    @classmethod
    def transform(cls, item):
        return int(item)


class ListOfFloat(List):
    @classmethod
    def transform(cls, item):
        return float(item)


class ListOfBool(List):
    @classmethod
    def transform(cls, item):
        return Bool.str2bool(item)


class Json(String):
    def __init__(self, key, **kwargs):
        if "default" in kwargs:
            kwargs["default"] = json.dumps(kwargs["default"])
        super(Json, self).__init__(key, **kwargs)

    def format(self, context):
        if self.value is not None:
            self.value = json.loads(self.value)
        return self.value

    def save(self, context, result):
        data = (
            result.value.to_dict()
            if isinstance(result.value, objects.Context)
            else result.value
        )
        self.logSaved(data)
        return json.dumps(data)



# coding=utf-8
from __future__ import print_function

from odps import ODPS
from odps.df import DataFrame

from suanpan.log import logger


class OdpsDataWarehouse(object):
    DEFAULT_ODPS_PROJECT_NAME = "None"

    DTYPE_SQLTYPE_MAPPING = {
        "int8": "tinyint",
        "int16": "smallint",
        "int32": "int",
        "int64": "bigint",
        "float32": "float",
        "float64": "double",
        "class": "string",
        "datetime": "timestamp",
    }

    def __init__(
        self, odpsAccessId, odpsAccessKey, odpsEndpoint, odpsProject, **kwargs
    ):
        self.odpsAccessId = odpsAccessId
        self.odpsAccessKey = odpsAccessKey
        self.odpsEndpoint = odpsEndpoint
        self.odpsProject = odpsProject

        self._connection = self.connect()

    def connect(self):
        connection = ODPS(
            access_id=self.odpsAccessId,
            secret_access_key=self.odpsAccessKey,
            endpoint=self.odpsEndpoint,
            project=self.odpsProject,
        )
        return connection

    def testConnection(self, connection):
        connection.execute_sql("select current_timestamp()")
        logger.info("Odps connected...")

    def readTable(self, table, partition=None, database=DEFAULT_ODPS_PROJECT_NAME):
        df = (
            DataFrame(self._connection.get_table(table).get_partition(partition))
            if partition
            else DataFrame(self._connection.get_table(table))
        )
        return df

    def writeTable(
        self,
        table,
        data,
        partition=None,
        database=DEFAULT_ODPS_PROJECT_NAME,
        overwrite=True,
    ):
        if not isinstance(data, DataFrame):
            data = DataFrame(data)

        if partition:
            data.persist(
                table,
                partition=partition,
                drop_partition=overwrite,
                create_partition=True,
                odps=self._connection,
            )
        else:
            data.persist(
                table,
                partition=partition,
                drop_table=overwrite,
                create_table=True,
                odps=self._connection,
            )

    def createTable(self, table, columns):
        self.execute(
            u"create table if not exists {table} ({columns})".format(
                table=table,
                columns=u",".join(
                    [
                        u"{} {}".format(column["name"], column["type"])
                        for column in columns
                    ]
                ),
            )
        )

    def dropTable(self, connection, table):
        connection.delete_table(table)

    def execute(self, sql):
        try:
            self._connection.execute_sql(sql)
        except Exception as e:
            logger.error("SQL: {}".format(sql))
            raise e

    def getColumns(self, data):
        return [
            {"name": name.split(".")[-1], "type": self.toSqlType(dtype)}
            for name, dtype in data.dtypes.items()
        ]

    def toSqlType(self, dtype):
        typeName = self.shortenDatetimeType(dtype.name)
        return self.DTYPE_SQLTYPE_MAPPING.get(typeName, "string")

    def shortenDatetimeType(self, typeName):
        datatimeType = "datetime"
        return datatimeType if typeName.startswith(datatimeType) else typeName



# coding=utf-8
from __future__ import print_function

import itertools

from suanpan.arguments import Bool, Int, String
from suanpan.mq.redis import RedisMQ
from suanpan.proxy import Proxy


class MQProxy(Proxy):
    MAPPING = {"redis": RedisMQ}
    DEFAULT_ARGUMENTS = [String("mq-type")]
    REDIS_ARGUMENTS = [
        String("mq-redis-host", default="localhost"),
        Int("mq-redis-port", default=6379),
        Bool("mq-redis-realtime", default=False),
    ]
    ARGUMENTS = list(itertools.chain(DEFAULT_ARGUMENTS, REDIS_ARGUMENTS))


mq = MQProxy()



# coding=utf-8
from __future__ import print_function

import itertools

from suanpan import interfaces, g
from suanpan.components import Component
from suanpan.dw import dw
from suanpan.log import logger
from suanpan.mstorage import mstorage
from suanpan.storage import storage


class DockerComponent(
    Component, interfaces.HasBaseServices, interfaces.HasLogger, interfaces.HasDevMode
):
    ENABLED_BASE_SERVICES = {"dw", "storage", "mstorage"}

    def initBase(self, args):
        logger.setLogger(self.name)
        self.setBaseServices(args)



# coding=utf-8
from __future__ import print_function

from contextlib import contextmanager

from pyspark.sql import SparkSession

from suanpan import interfaces, objects
from suanpan.components import Component
from suanpan.log import logger


class SparkComponent(Component, interfaces.HasLogger):
    @contextmanager
    def context(self, args):
        spark = (
            SparkSession.builder.appName(self.runFunc.__name__)
            .enableHiveSupport()
            .getOrCreate()
        )
        yield objects.Context(spark=spark)
        spark.stop()

    def initBase(self, args):
        logger.setLogger(self.name)



# coding=utf-8
from __future__ import print_function

from suanpan import arguments as base
from suanpan.dw import arguments as dw
from suanpan.mq import arguments as mq
from suanpan.mstorage import arguments as mstorage
from suanpan.storage import arguments as storage

Int = base.Int
String = base.String
Float = base.Float
Bool = base.Bool
List = mq.List
Json = base.Json
File = storage.File
Folder = storage.Folder
Npy = mstorage.Npy
Table = dw.Table



# coding=utf-8
from __future__ import print_function

import functools
import json

from suanpan import path

_load = functools.partial(json.load, encoding="utf-8")
_loads = functools.partial(json.loads, encoding="utf-8")
_dump = functools.partial(json.dump, ensure_ascii=False, encoding="utf-8")
_dumps = functools.partial(json.dumps, ensure_ascii=False)


def _loadf(file, *args, **kwargs):
    encoding = kwargs.pop("encoding", "utf-8")
    with open(file, "w", encoding=encoding) as file:
        return _load(file, *args, **kwargs)


def _dumpf(s, file, *args, **kwargs):
    encoding = kwargs.pop("encoding", "utf-8")
    path.safeMkdirsForFile(file)
    with open(file, "w", encoding=encoding) as file:
        return _dump(s, file, *args, **kwargs)


def load(file, *args, **kwargs):
    _l = _loadf if isinstance(file, str) else _load
    return _l(file, *args, **kwargs)


def dump(s, file, *args, **kwargs):
    _d = _dumpf if isinstance(file, str) else _dump
    return _d(s, file, *args, **kwargs)


loads = _loads
dumps = _dumps



# coding=utf-8
from __future__ import print_function

import sys
from datetime import datetime, timedelta

from colorama import init

try:
    from itertools import zip_longest as zip_longest
except:
    from itertools import izip_longest as zip_longest


VERSION = "0.2.0"

init()

# ANSI escape SGR Parameters color codes
AVAILABLE_COLORS = {
    "red": 91,
    "blue": 94,
    "green": 92,
    "magenta": 95,
    "yellow": 93,
    "black": 90,
    "cyan": 96,
}

DAYS = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]
DELIM = ","
TICK = "▇"
SM_TICK = "▏"

try:
    range = xrange
except NameError:
    pass


def find_min(list_):
    """Return the minimum value in sublist of list."""
    return min([sublist[-1] for sublist in list_])


def find_max(list_):
    """Return the maximum value in sublist of list."""
    return max([sublist[-1] for sublist in list_])


def find_max_label_length(labels):
    """Return the maximum length for the labels."""
    length = 0
    for i in range(len(labels)):
        if len(labels[i]) > length:
            length = len(labels[i])

    return length


def normalize(data, width):
    """Normalize the data and return it."""
    min_dat = find_min(data)
    # We offset by the minimum if there's a negative.
    off_data = []
    if min_dat < 0:
        min_dat = abs(min_dat)
        for dat in data:
            off_data.append([_d + min_dat for _d in dat])
    else:
        off_data = data
    min_dat = find_min(off_data)
    max_dat = find_max(off_data)

    if max_dat < width:
        # Don't need to normalize if the max value
        # is less than the width we allow.
        return off_data

    # max_dat / width is the value for a single tick. norm_factor is the
    # inverse of this value
    # If you divide a number to the value of single tick, you will find how
    # many ticks it does contain basically.
    norm_factor = width / float(max_dat)
    normal_dat = []
    for dat in off_data:
        normal_dat.append([_v * norm_factor for _v in dat])

    return normal_dat


def horiz_rows(labels, data, normal_dat, args, colors):
    """Prepare the horizontal graph.
       Each row is printed through the print_row function."""
    val_min = find_min(data)

    for i in range(len(labels)):
        if args["no_labels"]:
            # Hide the labels.
            label = ""
        else:
            label = "{:<{x}}: ".format(labels[i], x=find_max_label_length(labels))

        values = data[i]
        num_blocks = normal_dat[i]

        for j in range(len(values)):
            # In Multiple series graph 1st category has label at the beginning,
            # whereas the rest categories have only spaces.
            if j > 0:
                len_label = len(label)
                label = " " * len_label
            tail = " {}{}".format(args["format"].format(values[j]), args["suffix"])
            if colors:
                color = colors[j]
            else:
                color = None

            if not args["vertical"]:
                print(label, end="")

            yield (values[j], int(num_blocks[j]), val_min, color)

            if not args["vertical"]:
                print(tail)


# Prints a row of the horizontal graph.
def print_row(value, num_blocks, val_min, color):
    """A method to print a row for a horizontal graphs.

    i.e:
    1: ▇▇ 2
    2: ▇▇▇ 3
    3: ▇▇▇▇ 4
    """
    if color:
        sys.stdout.write("\033[{}m".format(color))  # Start to write colorized.

    if num_blocks < 1 and (value > val_min or value > 0):
        # Print something if it's not the smallest
        # and the normal value is less than one.
        sys.stdout.write(SM_TICK)
    else:
        for _ in range(num_blocks):
            sys.stdout.write(TICK)

    if color:
        sys.stdout.write("\033[0m")  # Back to original.


def stacked_graph(labels, data, normal_data, len_categories, args, colors):
    """Prepare the horizontal stacked graph.
       Each row is printed through the print_row function."""
    val_min = find_min(data)

    for i in range(len(labels)):
        if args["no_labels"]:
            # Hide the labels.
            label = ""
        else:
            label = "{:<{x}}: ".format(labels[i], x=find_max_label_length(labels))

        print(label, end="")
        values = data[i]
        num_blocks = normal_data[i]

        for j in range(len(values)):
            print_row(values[j], int(num_blocks[j]), val_min, colors[j])

        tail = " {}{}".format(args["format"].format(sum(values)), args["suffix"])
        print(tail)


value_list, zipped_list, vertical_list, maxi = [], [], [], 0


def vertically(value, num_blocks, val_min, color, args):
    """Prepare the vertical graph.
       The whole graph is printed through the print_vertical function."""
    global maxi, value_list

    value_list.append(str(value))

    # In case the number of blocks at the end of the normalization is less
    # than the default number, use the maxi variable to escape.
    if maxi < num_blocks:
        maxi = num_blocks

    if num_blocks > 0:
        vertical_list.append((TICK * num_blocks))
    else:
        vertical_list.append(SM_TICK)

    # Zip_longest method in order to turn them vertically.
    for row in zip_longest(*vertical_list, fillvalue="  "):
        zipped_list.append(row)

    counter, result_list = 0, []

    # Combined with the maxi variable, escapes the appending method at
    # the correct point or the default one (width).
    for i in reversed(zipped_list):
        result_list.append(i)
        counter += 1

        if maxi == args["width"]:
            if counter == (args["width"]):
                break
        else:
            if counter == maxi:
                break

    # Return a list of rows which will be used to print the result vertically.
    return result_list


def print_vertical(vertical_rows, labels, color, args):
    """Print the whole vertical graph."""
    if color:
        sys.stdout.write("\033[{}m".format(color))  # Start to write colorized.

    for row in vertical_rows:
        print(*row)

    sys.stdout.write("\033[0m")  # End of printing colored

    print("-" * len(row) + "Values" + "-" * len(row))
    # Print Values
    for value in zip_longest(*value_list, fillvalue=" "):
        print("  ".join(value))

    if args["no_labels"] == False:
        print("-" * len(row) + "Labels" + "-" * len(row))
        # Print Labels
        for label in zip_longest(*labels, fillvalue=""):
            print("  ".join(label))


def chart(colors, data, args, labels):
    """Handle the normalization of data and the printing of the graph."""
    len_categories = len(data[0])
    if len_categories > 1:
        # Stacked graph
        if args["stacked"]:
            normal_dat = normalize(data, args["width"])
            stacked_graph(labels, data, normal_dat, len_categories, args, colors)
            return

        if not colors:
            colors = [None] * len_categories

        # Multiple series graph with different scales
        # Normalization per category
        if args["different_scale"]:
            for i in range(len_categories):
                cat_data = []
                for dat in data:
                    cat_data.append([dat[i]])

                # Normalize data, handle negatives.
                normal_cat_data = normalize(cat_data, args["width"])

                # Generate data for a row.
                for row in horiz_rows(
                    labels, cat_data, normal_cat_data, args, [colors[i]]
                ):
                    # Print the row
                    if not args["vertical"]:
                        print_row(*row)
                    else:
                        vertic = vertically(*row, args=args)

                # Vertical graph
                if args["vertical"]:
                    print_vertical(vertic, labels, colors[i], args)

                print()
                value_list.clear(), zipped_list.clear(), vertical_list.clear()
                return

    # One category/Multiple series graph with same scale
    # All-together normalization
    if not args["stacked"]:
        normal_dat = normalize(data, args["width"])
        for row in horiz_rows(labels, data, normal_dat, args, colors):
            if not args["vertical"]:
                print_row(*row)
            else:
                vertic = vertically(*row, args=args)

        if args["vertical"] and len_categories == 1:
            if colors:
                color = colors[0]
            else:
                color = None

            print_vertical(vertic, labels, color, args)

        print()


def check_data(labels, data, args):
    """Check that all data were inserted correctly. Return the colors."""
    len_categories = len(data[0])

    # Check that there are data for all labels.
    if len(labels) != len(data):
        print(">> Error: Label and data array sizes don't match")
        sys.exit(1)

    # Check that there are data for all categories per label.
    for dat in data:
        if len(dat) != len_categories:
            print(">> Error: There are missing values")
            sys.exit(1)

    colors = []

    # If user inserts colors, they should be as many as the categories.
    if args["color"] is not None:
        if len(args["color"]) != len_categories:
            print(">> Error: Color and category array sizes don't match")
            sys.exit(1)

        for color in args["color"]:
            colors.append(AVAILABLE_COLORS.get(color))

    # Vertical graph for multiple series of same scale is not supported yet.
    if args["vertical"] and len_categories > 1 and not args["different_scale"]:
        print(
            ">> Error: Vertical graph for multiple series of same "
            "scale is not supported yet."
        )
        sys.exit(1)

    # If user hasn't inserted colors, pick the first n colors
    # from the dict (n = number of categories).
    if args["stacked"] and not colors:
        colors = [v for v in list(AVAILABLE_COLORS.values())[:len_categories]]

    return colors


def print_categories(categories, colors):
    """Print a tick and the category's name for each category above
       the graph."""
    for i in range(len(categories)):
        if colors:
            sys.stdout.write("\033[{}m".format(colors[i]))  # Start to write colorized.

        sys.stdout.write(TICK + " " + categories[i] + "  ")
        if colors:
            sys.stdout.write("\033[0m")  # Back to original.

    print("\n\n")


def read_data(args):
    """Read data from a file or stdin and returns it.

       Filename includes (categories), labels and data.
       We append categories and labels to lists.
       Data are inserted to a list of lists due to the categories.

       i.e.
       labels = ['2001', '2002', '2003', ...]
       categories = ['boys', 'girls']
       data = [ [20.4, 40.5], [30.7, 100.0], ...]"""
    filename = args["filename"]
    stdin = filename == "-"

    if args["verbose"]:
        print(">> Reading data from {}".format("stdin" if stdin else filename))

    print("")
    if args["title"]:
        print("# " + args["title"] + "\n")

    categories, labels, data, colors = ([] for i in range(4))

    f = sys.stdin if stdin else open(filename, "r")
    for line in f:
        line = line.strip()
        if line:
            if not line.startswith("#"):
                if line.find(DELIM) > 0:
                    cols = line.split(DELIM)
                else:
                    cols = line.split()

                # Line contains categories.
                if line.startswith("@"):
                    cols[0] = cols[0].replace("@ ", "")
                    categories = cols

                # Line contains label and values.
                else:
                    labels.append(cols[0].strip())
                    data_points = []
                    for i in range(1, len(cols)):
                        data_points.append(float(cols[i].strip()))

                    data.append(data_points)
    f.close()

    # Check that all data are valid. (i.e. There are no missing values.)
    colors = check_data(labels, data, args)
    if categories:
        # Print categories' names above the graph.
        print_categories(categories, colors)

    return categories, labels, data, colors


def calendar_heatmap(data, labels, args):
    """Print a calendar heatmap."""
    if args["color"]:
        colornum = AVAILABLE_COLORS.get(args["color"][0])
    else:
        colornum = AVAILABLE_COLORS.get("blue")

    dt_dict = {}
    for i in range(len(labels)):
        dt_dict[labels[i]] = data[i][0]

    # get max value
    max_val = float(max(data)[0])

    tick_1 = "░"
    tick_2 = "▒"
    tick_3 = "▓"
    tick_4 = "█"

    if args["custom_tick"]:
        tick_1 = tick_2 = tick_3 = tick_4 = args["custom_tick"]

    # check if start day set, otherwise use one year ago
    if args["start_dt"]:
        start_dt = datetime.strptime(args["start_dt"], "%Y-%m-%d")
    else:
        start = datetime.now()
        start_dt = datetime(year=start.year - 1, month=start.month, day=start.day)

    # modify start date to be a Monday, subtract weekday() from day
    start_dt = start_dt - timedelta(start_dt.weekday())

    # TODO: legend doesn't line up properly for all start dates/data
    # top legend for months
    sys.stdout.write("     ")
    for month in range(13):
        month_dt = datetime(
            year=start_dt.year, month=start_dt.month, day=1
        ) + timedelta(days=month * 31)
        sys.stdout.write(month_dt.strftime("%b") + " ")
        if args["custom_tick"]:  # assume custom tick is emoji which is one wider
            sys.stdout.write(" ")

    sys.stdout.write("\n")

    for day in range(7):
        sys.stdout.write(DAYS[day] + ": ")
        for week in range(53):
            day_ = start_dt + timedelta(days=day + week * 7)
            day_str = day_.strftime("%Y-%m-%d")

            if day_str in dt_dict:
                if dt_dict[day_str] > max_val * 0.75:
                    tick = tick_4
                elif dt_dict[day_str] > max_val * 0.50:
                    tick = tick_3
                elif dt_dict[day_str] > max_val * 0.25:
                    tick = tick_2
                else:
                    tick = tick_1
            else:
                tick = " "

            if colornum:
                sys.stdout.write("\033[{}m".format(colornum))

            sys.stdout.write(tick)
            if colornum:
                sys.stdout.write("\033[0m")

        sys.stdout.write("\n")



# coding=utf-8
from __future__ import print_function

import functools
import itertools
import time

from suanpan.log import logger


def formatFuncCall(func, *args, **kwargs):
    paramString = ", ".join(
        itertools.chain(
            (str(a) for a in args), ("{}={}".format(k, v) for k, v in kwargs.items())
        )
    )
    funcString = "{}({})".format(func.__name__, paramString)
    return funcString


def costCall(func, *args, **kwargs):
    startTime = time.time()
    result = func(*args, **kwargs)
    endTime = time.time()
    costTime = endTime - startTime
    return costTime, result


def cost(func):
    @functools.wraps(func)
    def _dec(*args, **kwargs):
        costTime, result = costCall(func, *args, **kwargs)
        logger.info("{} - {}s".format(formatFuncCall(func, *args, **kwargs), costTime))
        return result

    return _dec


def print(func):
    @functools.wraps(func)
    def _dec(*args, **kwargs):
        result = func(*args, **kwargs)
        logger.info("{} - {}".format(formatFuncCall(func, *args, **kwargs), result))
        return result

    return _dec
